<?php

declare(strict_types=1);

namespace Drupal\field_updater_service\Commands;

use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\field_updater_service\FieldUpdater;
use Drush\Commands\DrushCommands;
use Symfony\Component\Console\Helper\Table;
use Symfony\Component\Console\Question\Question;
use Symfony\Component\Console\Style\SymfonyStyle;

/**
 * Drush commands for Field Updater.
 */
class FieldUpdaterCommands extends DrushCommands {

  use StringTranslationTrait;

  /**
   * The field updater service.
   *
   * @var \Drupal\field_updater_service\FieldUpdater
   */
  private readonly FieldUpdater $updaterService;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  private readonly EntityTypeManagerInterface $entityTypeManager;

  /**
   * Constructs a new FieldUpdaterCommands object.
   */
  public function __construct(FieldUpdater $updater_service, EntityTypeManagerInterface $entity_type_manager) {
    $this->updaterService = $updater_service;
    $this->entityTypeManager = $entity_type_manager;
    parent::__construct();
  }

  /**
   * Updates entity fields based on a FieldUpdaterType config entity.
   *
   * @command field-updater:update
   * @aliases fu-update
   * @usage drush field-updater:update my_updater_id
   * @description Updates entities based on a FieldUpdaterType config entity in batches.
   */
  public function update(): void {
    $id = $this->promptStorage();
    // Confirmation.
    $io = new SymfonyStyle($this->input, $this->output);
    if (!$io->confirm('Proceed with running this update?')) {
      $io->warning('Operation cancelled.');
      return;
    }
    try {
      $this->updaterService->updateEntities($id);
    }
    catch (InvalidPluginDefinitionException | PluginNotFoundException $e) {
      // No valid plugin found so abort.
    }
    $batch =& batch_get();
    if (!empty($batch)) {
      drush_backend_batch_process();
    }
    $this->io->success((string) $this->t('Successfully ran updater @id', ['@id' => $id]));

    // Output example update hook.
    $this->outputUpdateHookExample($id);

    $this->logger()->success(dt('Initiated batch update for @id.', ['@id' => $id]));
  }

  /**
   * Lists all FieldUpdaterType config entity IDs.
   *
   * @command field-updater:list
   * @aliases fu-list
   * @usage drush field-updater:list
   * @description Displays a list of all FieldUpdaterType config entity IDs.
   */
  public function listUpdaters(): void {
    $storage = $this->entityTypeManager->getStorage('field_updater_type');
    $ids = $storage->getQuery()
      ->accessCheck(FALSE)
      ->execute();

    if (empty($ids)) {
      $this->logger()->warning(dt('No FieldUpdaterType entities found.'));
      return;
    }

    $entities = $storage->loadMultiple($ids);
    $output = [];

    foreach ($entities as $entity) {
      $output[] = [
        'ID' => $entity->id(),
        'Label' => $entity->label(),
      ];
    }

    // Output as a table.
    $this->output()->writeln(dt('FieldUpdaterType Entities:'));
    $this->io()->table(
      ['ID', 'Label'],
      $output
    );

    $this->logger()->success(dt('Listed @count FieldUpdaterType entities.', ['@count' => count($ids)]));
  }

  /**
   * Prompt for the updater config entity id.
   *
   * @return string
   *   The entity id of the updater.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function promptStorage(): string {
    $storage = $this->entityTypeManager->getStorage('field_updater_type');
    $ids = $storage->getQuery()
      ->accessCheck(FALSE)
      ->execute();

    if (empty($ids)) {
      throw new \RuntimeException('No Field updater entities found.');
    }

    /** @var \Drupal\field_updater_service\Entity\FieldUpdaterType[] $entities */
    $entities = $storage->loadMultiple($ids);
    $choices = [];
    $rows = [];
    $index = 0;

    foreach ($entities as $entity) {
      $rows[] = [
        $index,
        $entity->id(),
        $entity->label(),
      ];
      $choices[$index] = $entity->id();
      $index++;
    }

    // Display the table.
    $table = new Table($this->output());
    $table
      ->setHeaders(['#', 'Id', 'Label'])
      ->setRows($rows);
    $table->render();

    $prompt = $this->t('Enter the number (#) of the field updater instance');
    $question = new Question((string) $prompt);
    $question->setValidator(function ($answer) use ($choices) {
      if (!isset($choices[$answer])) {
        throw new \RuntimeException('Invalid selection: ' . $answer);
      }
      return $choices[$answer];
    });

    return $this->io()->askQuestion($question);
  }

  /**
   * Output an example update hook for the executed command.
   *
   * @param string $id
   *   The entity id of the updater.
   */
  protected function outputUpdateHookExample(string $id): void {
    $io = $this->io();
    $io->section('For deploying this update to other environments, copy the example update hook to your custom module .install file. Replace the module name (MY_MODULE) and sequence (N) accordingly.');
    $hook_name = 'MY_MODULE_update_N';
    $description = "Runs the FieldUpdaterService for the '{$id}' config entity.";

    // Start the code block.
    $code[] = "/**";
    $code[] = "* {$description}";
    $code[] = "*/";
    $code[] = "function $hook_name() {";
    $code[] = "  \\Drupal::service('field_updater_service.updater')->updateEntities('$id');";
    $code[] = "}";
    $code[] = "";

    // Output the formatted code block.
    $io->block(implode("\n", $code), NULL, 'fg=green;bg=black', ' ', TRUE);
  }

}
