<?php

namespace Drupal\field_updater_service\Entity;

use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\field_updater_service\FieldUpdaterTypeInterface;

/**
 * Defines the Field Updater Type configuration entity.
 *
 * @ConfigEntityType(
 *   id = "field_updater_type",
 *   label = @Translation("Field updater type"),
 *   label_collection = @Translation("Field updater types"),
 *   handlers = {
 *     "form" = {
 *       "default" = "Drupal\field_updater_service\Form\FieldUpdaterForm",
 *       "add" = "Drupal\field_updater_service\Form\FieldUpdaterForm",
 *       "edit" = "Drupal\field_updater_service\Form\FieldUpdaterForm",
 *       "delete" = "Drupal\Core\Entity\EntityDeleteForm",
 *     },
 *     "list_builder" = "Drupal\field_updater_service\FieldUpdaterTypeListBuilder",
 *     "route_provider" = {
 *       "html" = "Drupal\Core\Entity\Routing\AdminHtmlRouteProvider",
 *     },
 *   },
 *   config_prefix = "field_updater_type",
 *   admin_permission = "administer site configuration",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "label",
 *     "uuid" = "uuid",
 *     "status" = "status",
 *   },
 *   links = {
 *     "canonical" = "/admin/config/field-updater/{field_updater_type}",
 *     "add-form" = "/admin/config/field-updater/add",
 *     "edit-form" = "/admin/config/field-updater/{field_updater_type}/edit",
 *     "mapping" = "/admin/config/field-updater/{field_updater_type}/mapping",
 *     "delete-form" = "/admin/config/field-updater/{field_updater_type}/delete",
 *     "collection" = "/admin/config/field-updater",
 *   },
 *   config_export = {
 *     "id",
 *     "label",
 *     "entity_type",
 *     "bundle",
 *     "field_target",
 *     "field_source",
 *     "target_type",
 *     "target_bundles",
 *     "mappings",
 *   }
 * )
 */
class FieldUpdaterType extends ConfigEntityBase implements FieldUpdaterTypeInterface {

  /**
   * The entity type.
   *
   * @var string
   */
  public $entity_type;

  /**
   * The bundle.
   *
   * @var string
   */
  public $bundle;

  /**
   * The target field machine name.
   *
   * @var string
   */
  public $field_target;

  /**
   * The source field machine name.
   *
   * @var string
   */
  public $field_source;

  /**
   * The target type machine name.
   *
   * @var string
   */
  public $target_type;

  /**
   * The target bundle machine names.
   *
   * @var array
   */
  public array $target_bundles = [];

  /**
   * The list of source to target mappings.
   *
   * @var array
   */
  public array $mappings = [];

  /**
   * {@inheritdoc}
   */
  public function getMappings() {
    return $this->mappings;
  }

  /**
   * {@inheritdoc}
   */
  public function setMappings(array $mappings) {
    $this->mappings = $mappings;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function removeMappings() {
    $this->mappings = [];
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getUpdaterEntityTypeId(): string {
    return $this->entity_type;
  }

  /**
   * {@inheritdoc}
   */
  public function getUpdaterBundle(): string {
    return $this->bundle;
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    parent::calculateDependencies();
    /** @var \Drupal\Core\Entity\EntityFieldManagerInterface $field_manager */
    $field_manager = \Drupal::service('entity_field.manager');
    $entity_type_id = $this->entity_type;
    $bundle = $this->bundle;
    $field_target = $this->field_target;
    $field_source = $this->field_source;
    $mappings = $this->mappings;
    $mapping_definitions = [];
    if ($entity_type = $this->entityTypeManager()->getDefinition($entity_type_id)) {
      $this->addDependency('module', $entity_type->getProvider());

      // Add bundle dependency if applicable.
      if ($entity_type->hasKey('bundle')) {
        $bundle_entity_type_id = $entity_type->getBundleEntityType();
        if ($bundle_storage = $this->entityTypeManager()->getStorage($bundle_entity_type_id)) {
          $this->addDependency('config', $bundle_storage->load($bundle)
            ->getConfigDependencyName());
        }
      }
      $field_definitions = $field_manager->getFieldDefinitions($entity_type_id, $bundle);
      $mapping_definitions = $field_definitions;
      // Add dependency for the field target config.
      if (isset($field_definitions[$field_target])) {
        $this->addDependency('config', $field_definitions[$field_target]->getConfigDependencyName());
      }
      // Add dependency for the field source config.
      if (isset($field_definitions[$field_source])) {
        $this->addDependency('config', $field_definitions[$field_source]->getConfigDependencyName());
      }
    }
    // Add dependencies for the target entity type and bundle if applicable.
    if (!empty($this->target_type) && $target_type = $this->entityTypeManager()->getDefinition($this->target_type)) {
      $this->addDependency('module', $target_type->getProvider());
      // Add field storage dependencies.
      $mapping_definitions = $field_manager->getFieldStorageDefinitions($this->target_type);

      // Add bundle dependencies if applicable.
      if ($target_type->hasKey('bundle')) {
        $target_bundle_entity_type_id = $target_type->getBundleEntityType();
        if ($bundle_storage = $this->entityTypeManager()
          ->getStorage($target_bundle_entity_type_id)) {
          $target_bundles = !empty($this->target_bundles) ? $this->target_bundles : [];
          $loaded_bundles = $bundle_storage->loadMultiple($target_bundles);

          // Add bundle config dependencies.
          foreach ($loaded_bundles as $target_bundle) {
            $this->addDependency('config', $target_bundle->getConfigDependencyName());
          }
        }
      }
    }
    // Loop over the mappings and add field dependencies.
    foreach ($mappings as $mapping) {
      $source_field = $mapping_definitions[$mapping['source_field']] ?? NULL;
      if (empty($source_field)) {
        continue;
      }
      if ($source_field instanceof FieldDefinitionInterface) {
        $source_field = $source_field->getFieldStorageDefinition();
      }
      if ($source_field->isBaseField()) {
        continue;
      }
      $this->addDependency('config', $source_field->getConfigDependencyName());
    }

    return $this;
  }

}
