<?php

declare(strict_types=1);

namespace Drupal\field_updater_service;

use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Utility\Token;
use Drupal\field\Entity\FieldStorageConfig;

/**
 * Service to update entity fields based on FieldUpdaterType config entity.
 */
class FieldUpdater {

  use DependencySerializationTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The logger channel service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected LoggerChannelInterface $logger;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected MessengerInterface $messenger;

  /**
   * The token service.
   *
   * @var \Drupal\Core\Utility\Token
   */
  protected Token $tokenService;

  /**
   * Constructs a new FieldUpdaterService.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger channel for field_updater.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Core\Utility\Token $token_service
   *   The token service.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    LoggerChannelFactoryInterface $logger_factory,
    MessengerInterface $messenger,
    Token $token_service,
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->logger = $logger_factory->get('field_updater_service');
    $this->messenger = $messenger;
    $this->tokenService = $token_service;
  }

  /**
   * Initiates a batch process to update entity fields.
   *
   * @param string $id
   *   The field updater config id.
   * @param bool $use_batch
   *   Whether to use batch processing (default TRUE).
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function updateEntities(string $id, bool $use_batch = TRUE): void {
    /** @var \Drupal\field_updater_service\Entity\FieldUpdaterType $field_updater */
    $field_updater = $this->entityTypeManager->getStorage('field_updater_type')->load($id);
    if (!$field_updater) {
      $message = 'There is no field_updater type with the id: ' . $id . '.';
      throw new \InvalidArgumentException($message);
    }
    $entity_type_id = $field_updater->getUpdaterEntityTypeId();
    $bundle = $field_updater->getUpdaterBundle();
    $field_target = $field_updater->get('field_target');
    $field_source = $field_updater->get('field_source');
    $target_type = $field_updater->get('target_type');
    $target_bundles = $field_updater->get('target_bundles');
    $mappings = $field_updater->getMappings();

    if (empty($entity_type_id) || empty($bundle) || empty($field_target) || empty($mappings)) {
      $this->logger->warning('Incomplete configuration for @id: entity_type, bundle, field_target, or mappings missing.', [
        '@id' => $field_updater->id(),
      ]);
      return;
    }

    $target_field = FieldStorageConfig::loadByName($entity_type_id, $field_target);
    $cardinality = $target_field->getCardinality();

    // Query entity IDs.
    $storage = $this->entityTypeManager->getStorage($entity_type_id);
    $entity_type = $storage->getEntityType();
    $query = $storage->getQuery()
      ->condition($entity_type->getKey('bundle'), $bundle)
      ->accessCheck(FALSE);
    $entity_ids = $query->execute();

    if (empty($entity_ids)) {
      $this->logger->notice('No entities found for @type/@bundle.', [
        '@type' => $entity_type_id,
        '@bundle' => $bundle,
      ]);
      return;
    }

    if ($use_batch) {
      // Process in batches.
      $this->createBatch($field_updater, $entity_ids, $cardinality);
    }
    else {
      // Process synchronously (for testing or small datasets).
      /** @var \Drupal\Core\Entity\ContentEntityInterface[] $entities */
      $entities = $storage->loadMultiple($entity_ids);
      foreach ($entities as $entity) {
        $this->updateEntity($entity, $field_target, $field_source, $target_type, $target_bundles, $mappings, $cardinality);
      }
      $this->logger->info('Processed @count entities for @type/@bundle.', [
        '@count' => count($entities),
        '@type' => $entity_type_id,
        '@bundle' => $bundle,
      ]);
    }
  }

  /**
   * Creates a batch process for updating entities.
   *
   * @param \Drupal\field_updater_service\FieldUpdaterTypeInterface $field_updater
   *   The field updater config entity.
   * @param array $entity_ids
   *   Array of entity IDs to process.
   * @param int $cardinality
   *   The maximum number of items allowed for the target field.
   */
  protected function createBatch(FieldUpdaterTypeInterface $field_updater, array $entity_ids, int $cardinality): void {
    $batch_builder = (new BatchBuilder())
      ->setTitle(t('Updating entities for @id', ['@id' => $field_updater->id()]))
      ->setFinishCallback([$this, 'batchFinished']);

    $entity_type_id = $field_updater->getUpdaterEntityTypeId();
    $field_target = $field_updater->get('field_target');
    $field_source = $field_updater->get('field_source') ?? NULL;
    $target_type = $field_updater->get('target_type') ?? NULL;
    $target_bundles = $field_updater->get('target_bundles') ?? [];
    $mappings = $field_updater->getMappings();
    $batch_size = 10;
    $chunks = array_chunk($entity_ids, $batch_size);

    foreach ($chunks as $chunk) {
      $batch_builder->addOperation([$this, 'batchProcess'], [
        $this,
        $entity_type_id,
        $field_target,
        $field_source,
        $target_type,
        $target_bundles,
        $mappings,
        $cardinality,
        $chunk,
      ]);
    }

    batch_set($batch_builder->toArray());
  }

  /**
   * Batch operation callback to process a chunk of entities.
   */
  public static function batchProcess(self $service, string $entity_type_id, string $field_target, ?string $field_source, ?string $target_type, array $target_bundles, array $mappings, int $cardinality, array $entity_ids, &$context): void {
    $storage = $service->entityTypeManager->getStorage($entity_type_id);
    /** @var \Drupal\Core\Entity\ContentEntityInterface[] $entities */
    $entities = $storage->loadMultiple($entity_ids);

    $updated_count = 0;
    foreach ($entities as $entity) {
      if ($service->updateEntity($entity, $field_target, $field_source, $target_type, $target_bundles, $mappings, $cardinality)) {
        $updated_count++;
      }
    }

    // Update batch context for progress.
    if (!isset($context['results']['processed'])) {
      $context['results']['processed'] = 0;
      $context['results']['updated'] = 0;
    }
    $context['results']['processed'] += count($entities);
    $context['results']['updated'] += $updated_count;

    $context['message'] = t('Processed @count entities.', ['@count' => $context['results']['processed']]);
  }

  /**
   * Batch finished callback.
   */
  public static function batchFinished($success, $results, $operations): void {
    if ($success) {
      $message = t('Finished processing @processed entities; @updated were updated.', [
        '@processed' => $results['processed'] ?? 0,
        '@updated' => $results['updated'] ?? 0,
      ]);
      \Drupal::messenger()->addStatus($message);
      \Drupal::logger('field_updater')->info($message);
    }
    else {
      $message = t('Batch process failed.');
      \Drupal::messenger()->addError($message);
      \Drupal::logger('field_updater')->error($message);
    }
  }

  /**
   * Updates a single entity based on mappings.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity to update.
   * @param string $field_target
   *   The target field name.
   * @param string|null $field_source
   *   The source field name if applicable (for referenced entities).
   * @param string|null $target_type
   *   The target entity type name if applicable (for referenced entities).
   * @param array $target_bundles
   *   The target bundle names if applicable (for referenced entities).
   * @param array $mappings
   *   The mappings array from the config entity.
   * @param int $cardinality
   *   The maximum number of items allowed for target field (-1 for unlimited).
   *
   * @return bool
   *   TRUE if the entity was updated, FALSE otherwise.
   */
  public function updateEntity(ContentEntityInterface $entity, string $field_target, ?string $field_source, ?string $target_type, array $target_bundles, array $mappings, int $cardinality): bool {
    if (!$entity->hasField($field_target)) {
      $this->logger->warning('Target field @field not found on entity @id.', [
        '@field' => $field_target,
        '@id' => $entity->id(),
      ]);
      return FALSE;
    }

    $changed = FALSE;
    $languages = $entity->getTranslationLanguages();

    foreach ($languages as $langcode => $language) {
      if (!$entity->hasTranslation($langcode)) {
        continue;
      }
      $translated_entity = $entity->getTranslation($langcode);
      $target_field = $translated_entity->get($field_target);
      $new_values = [];

      if (!empty($field_source)) {
        if (!$translated_entity->hasField($field_source)) {
          $this->logger->warning('Source field @field not found on entity @id.', [
            '@field' => $field_source,
            '@id' => $entity->id(),
          ]);
          return FALSE;
        }

        $field_definition = $translated_entity->getFieldDefinition($field_source);
        $reference_types = [
          'entity_reference_revisions',
          'entity_reference',
        ];
        if (in_array($field_definition->getType(), $reference_types)) {
          /** @var \Drupal\Core\Entity\ContentEntityInterface[] $referenced_entities */
          $referenced_entities = $translated_entity->get($field_source)->referencedEntities();
          $source_entities = [];
          if (!empty($referenced_entities)) {
            foreach ($referenced_entities as $ref_entity) {
              if ($target_type && $ref_entity->getEntityTypeId() !== $target_type) {
                $this->logger->warning('Referenced entity type @type does not match expected @expected for entity @id.', [
                  '@type' => $ref_entity->getEntityTypeId(),
                  '@expected' => $target_type,
                  '@id' => $entity->id(),
                ]);
                return FALSE;
              }
              if (!empty($target_bundles) && !in_array($ref_entity->bundle(), $target_bundles)) {
                $this->logger->warning('Referenced entity bundle @bundle does not match expected one of @expected for entity @id.', [
                  '@bundle' => $ref_entity->bundle(),
                  '@expected' => implode(', ', $target_bundles),
                  '@id' => $entity->id(),
                ]);
                return FALSE;
              }
              $source_entities[] = $ref_entity->hasTranslation($langcode) ? $ref_entity->getTranslation($langcode) : $ref_entity;
            }
          }
          $max_items = ($cardinality == -1 || $cardinality > 1) ? count($source_entities) : 1;

          foreach (array_slice($source_entities, 0, $max_items) as $delta => $source_entity) {
            $item_values = [];
            foreach ($mappings as $target_property => $mapping) {
              $source_value = $this->getSourceValue($source_entity, $mapping);
              // Skip invalid mappings.
              if ($source_value === FALSE) {
                continue;
              }

              $target_property_definition = $target_field->getFieldDefinition()
                ->getFieldStorageDefinition()
                ->getPropertyDefinition($target_property);
              if ($target_property_definition) {
                $item_values[$target_property] = $source_value;
              }
              else {
                $this->logger->warning('Target property @prop not found on field @field for entity @id.', [
                  '@prop' => $target_property,
                  '@field' => $field_target,
                  '@id' => $entity->id(),
                ]);
              }
            }
            if (!empty($item_values)) {
              $new_values[$delta] = $item_values;
            }
          }
        }
      }
      else {
        $temp_values = [];
        $max_count = 0;
        // Track whether each target_property comes from '_default_value',
        // '_default_null', or a field.
        $source_types = [];
        foreach ($mappings as $target_property => $mapping) {
          $source_field = $mapping['source_field'] ?? NULL;
          $source_property = $mapping['source_property'] ?? NULL;
          if (empty($source_field)) {
            continue;
          }
          $temp_values[$target_property] = [];
          $source_types[$target_property] = $source_field;
          if ($source_field === '_default_value') {
            $entity_type_id = $translated_entity->getEntityTypeId();
            $source_property_value = $this->tokenService->replace($source_property, [$entity_type_id => $translated_entity], [
              'clear' => TRUE,
              'langcode' => $langcode,
            ]);
            $temp_values[$target_property][] = $source_property_value;
          }
          elseif ($source_field === '_default_null') {
            $temp_values[$target_property][] = NULL;
          }
          elseif ($translated_entity->hasField($source_field)) {
            $field_values = $translated_entity->get($source_field)->getValue();
            $count = count($field_values);
            $limit = ($cardinality == -1) ? $count : min($count, $cardinality);
            for ($delta = 0; $delta < $limit; $delta++) {
              $temp_values[$target_property][] = $field_values[$delta][$source_property] ?? NULL;
            }
          }

          // Update max_count based on the number of values for this property.
          $max_count = max($max_count, count($temp_values[$target_property]));
        }
        // Build $new_values with consistent count.
        for ($delta = 0; $delta < $max_count; $delta++) {
          $item = [];
          foreach ($temp_values as $target_property => $values) {
            // If source is '_default_value' or '_default_null', repeat the last
            // value.
            if (in_array($source_types[$target_property], ['_default_value', '_default_null'])) {
              $item[$target_property] = $values[$delta] ?? ($values ? end($values) : NULL);
            }
            // For regular fields, use NULL for missing values.
            else {
              $item[$target_property] = $values[$delta] ?? NULL;
            }
          }
          $new_values[] = $item;
        }
      }

      // Set values and check for changes.
      if (!empty($new_values)) {
        $current_values = $target_field->getValue();
        if ($current_values !== $new_values) {
          $target_field->setValue($new_values);
          $changed = TRUE;
        }
      }
    }

    if ($changed) {
      try {
        $entity->save();
        return TRUE;
      }
      catch (\Exception $e) {
        $this->logger->error('Failed to save entity @id: @message', [
          '@id' => $entity->id(),
          '@message' => $e->getMessage(),
        ]);
      }
    }
    return FALSE;
  }

  /**
   * Extracts a source value from an entity based on a mapping.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $source_entity
   *   The source entity.
   * @param array $mapping
   *   The mapping configuration.
   *
   * @return mixed|false
   *   The source value, or FALSE if invalid.
   */
  protected function getSourceValue(ContentEntityInterface $source_entity, array $mapping): mixed {
    $source_field = $mapping['source_field'] ?? NULL;
    $source_property = $mapping['source_property'] ?? NULL;

    if (empty($source_field)) {
      return FALSE;
    }

    if ($source_field === '_default_value') {
      $source_value = $source_property;
      return ($source_value === '' || $source_value === NULL) ? FALSE : $source_value;
    }

    if ($source_field === '_default_null') {
      return NULL;
    }

    if (!$source_entity->hasField($source_field)) {
      $this->logger->warning('Source field @field not found on entity @id.', [
        '@field' => $source_field,
        '@id' => $source_entity->id(),
      ]);
      return FALSE;
    }

    $source_field_value = $source_entity->get($source_field)->getValue();
    return !empty($source_field_value) ? ($source_field_value[0][$source_property] ?? NULL) : NULL;
  }

}
