<?php

declare(strict_types=1);

namespace Drupal\field_updater_service\Form;

use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\field_updater_service\FieldUpdaterTypeInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form for Field updater type entities.
 */
class FieldUpdaterForm extends EntityForm {

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected EntityFieldManagerInterface $entityFieldManager;

  /**
   * The entity type bundle info.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected EntityTypeBundleInfoInterface $entityTypeBundleInfo;

  /**
   * Constructs a FieldMigrateMappingForm object.
   */
  public function __construct(EntityFieldManagerInterface $entity_field_manager, EntityTypeManagerInterface $entity_type_manager, EntityTypeBundleInfoInterface $entity_type_bundle_info) {
    $this->entityFieldManager = $entity_field_manager;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_field.manager'),
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state): array {
    $form = parent::form($form, $form_state);

    /** @var \Drupal\field_updater_service\Entity\FieldUpdaterType $entity */
    $entity = $this->entity;

    // Label field.
    $form['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Label'),
      '#default_value' => $entity->label(),
      '#required' => TRUE,
    ];

    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $entity->id(),
      '#machine_name' => [
        'exists' => [$this, 'exist'],
        'source' => ['label'],
      ],
      '#disabled' => !$entity->isNew(),
    ];

    // Entity type select field.
    $entity_types = $this->getEntityTypesWithBundles();
    $options = array_map(function ($entity_type) {
      return $entity_type['label'];
    }, $entity_types);

    $form['entity_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Entity type'),
      '#options' => $options,
      '#empty_option' => $this->t('- Select -'),
      '#default_value' => $entity->entity_type,
      '#required' => TRUE,
      '#ajax' => [
        'callback' => [$this, 'updateFields'],
        'wrapper' => 'fields-wrapper',
        'event' => 'change',
      ],
    ];

    // Wrapper for AJAX-updated fields.
    $form['fields_wrapper'] = [
      '#type' => 'container',
      '#attributes' => ['id' => 'fields-wrapper'],
    ];

    // Get entity type from form state or entity.
    $entity_type_id = $form_state->getValue('entity_type') ?: ($entity->entity_type ?? '');
    if (!empty($entity_type_id)) {
      // Bundle select field.
      $bundle_options = $entity_types[$entity_type_id]['bundles'] ?? [];
      $form['fields_wrapper']['bundle'] = [
        '#type' => 'select',
        '#title' => $this->t('Bundle'),
        '#options' => $bundle_options,
        '#empty_option' => $this->t('- Select -'),
        '#default_value' => $entity->bundle ?? '',
        '#required' => TRUE,
        '#ajax' => [
          'callback' => [$this, 'updateFields'],
          'wrapper' => 'fields-wrapper',
          'event' => 'change',
          'progress' => ['type' => 'throbber', 'message' => $this->t('Loading fields...')],
        ],
      ];

      // Get the bundle from form state or entity.
      $bundle = $form_state->getValue('bundle') ?: ($entity->bundle ?? '');
      if (!empty($bundle)) {
        // Get field definitions for the selected entity type.
        $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);
        $field_options = [];
        $reference_options = [];
        foreach ($field_definitions as $field_name => $definition) {
          if ($definition->getFieldStorageDefinition()->isBaseField() === FALSE) {
            $label = $this->t('@field_label (@field_name)', [
              '@field_label' => $definition->getLabel(),
              '@field_name' => $field_name,
            ]);
            if (!in_array($definition->getType(), [
              'entity_reference_revisions',
              'entity_reference',
            ])) {
              $field_options[$field_name] = $label;
            }
            else {
              $reference_options[$field_name] = $label;
            }
          }
        }
        $form['fields_wrapper']['field_target'] = [
          '#type' => 'select',
          '#title' => $this->t('Target field'),
          '#description' => $this->t('Select the target field to update values to.'),
          '#options' => $field_options,
          '#empty_option' => $this->t('- Select -'),
          '#default_value' => $entity->field_target,
          '#required' => TRUE,
          '#states' => [
            'visible' => [
              ':input[name="bundle"]' => ['!value' => ''],
            ],
          ],
        ];
        if (!empty($reference_options)) {
          $entity_source = !empty($entity->field_source) ? 'entity_reference' : 'entity';
          if (!empty($form_state->getValue('entity_source'))) {
            $entity_source = $form_state->getValue('entity_source');
          }
          $form['fields_wrapper']['entity_source'] = [
            '#title' => $this->t('Field mappings source'),
            '#type' => 'radios',
            '#options' => [
              'entity' => $this->t('Entity'),
              'entity_reference' => $this->t('Entity reference'),
            ],
            '#default_value' => $entity_source,
            '#ajax' => [
              'callback' => [$this, 'updateFields'],
              'wrapper' => 'fields-wrapper',
              'event' => 'change',
            ],
            '#states' => [
              'visible' => [
                ':input[name="bundle"]' => ['!value' => ''],
                ':input[name="field_target"]' => ['!value' => ''],
              ],
            ],
          ];

          if ($entity_source === 'entity_reference') {
            $form['fields_wrapper']['field_source'] = [
              '#type' => 'select',
              '#title' => $this->t('Entity reference'),
              '#description' => $this->t('Select an entity reference field as the source entity to map values from.'),
              '#options' => $reference_options,
              '#empty_option' => $this->t('- Select -'),
              '#default_value' => $entity->field_source,
              '#required' => TRUE,
              '#ajax' => [
                'callback' => [$this, 'updateFields'],
                'wrapper' => 'fields-wrapper',
                'event' => 'change',
              ],
              '#states' => [
                'visible' => [
                  ':input[name="field_target"]' => ['!value' => ''],
                ],
              ],
            ];
            $field_source = $form_state->getValue('field_source') ?: $entity->field_source;
            $target_type = NULL;
            $target_bundles = [];
            if (!empty($field_source)) {
              $source_definition = $field_definitions[(string) $field_source] ?? NULL;
              if ($source_definition) {
                $target_type = $source_definition->getFieldStorageDefinition()
                  ->getSetting('target_type');
                $entity_type = $this->entityTypeManager->getDefinition($target_type);
                $bundle_entity_type = $entity_type->getBundleEntityType();
                $all_bundles = $this->entityTypeManager->getStorage($bundle_entity_type)
                  ->loadMultiple();
                $target_bundle_options = array_map(function ($bundle) {
                  return $this->t('@label', ['@label' => $bundle->label()]);
                }, $all_bundles);

                // Get selected bundles and negate setting from source
                // definition.
                $selected_bundles = $source_definition->getSetting('handler_settings')['target_bundles'] ?? [];
                $negate = $source_definition->getSetting('handler_settings')['negate'] ?? FALSE;

                // Ensure $target_bundles keys are preserved (they are already
                // bundle IDs from $all_bundles).
                if ($negate) {
                  // Return bundles in $target_bundles whose keys are NOT in
                  // $selected_bundles.
                  $options = array_diff_key($target_bundle_options, $selected_bundles);
                }
                else {
                  // Return bundles in $target_bundles whose keys are in
                  // $selected_bundles.
                  $options = array_intersect_key($target_bundle_options, $selected_bundles);
                }
                // If for some reason no bundles are enabled, just return all.
                if (empty($options)) {
                  $options = $target_bundle_options;
                }
              }
              $target_bundles = $options;
            }
            $form['fields_wrapper']['target_type'] = [
              '#type' => 'value',
              '#value' => $target_type,
            ];
            if (empty($target_type) || empty($target_bundles)) {
              $form['fields_wrapper']['target_bundles'] = [
                '#type' => 'value',
                '#value' => [],
              ];
            }
            else {
              $form['fields_wrapper']['target_bundles'] = [
                '#type' => 'checkboxes',
                '#title' => $this->t('Entity reference @type', [
                  '@target_type' => $target_type,
                  '@type' => $this->formatPlural(count($target_bundles), 'bundle', 'bundles'),
                ]),
                '#description' => $this->t('Select the @target_type @type to map values from.', [
                  '@target_type' => $target_type,
                  '@type' => $this->formatPlural(count($target_bundles), 'bundle', 'bundles'),
                ]),
                '#options' => $target_bundles,
                '#default_value' => $entity->target_bundles ?? [],
                '#required' => TRUE,
                '#states' => [
                  'visible' => [
                    ':input[name="field_target"]' => ['!value' => ''],
                  ],
                ],
              ];
            }
          }
        }
      }
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state): array {
    $actions = parent::actions($form, $form_state);

    if ($this->entity->isNew()) {
      $actions['submit']['#value'] = $this->t('Save and add mappings');
    }
    else {
      $actions['submit']['#value'] = $this->t('Save updater type');
    }
    $actions['submit']['#submit'][] = '::toMapping';

    return $actions;
  }

  /**
   * Submit for new updaters to redirect to mappings.
   *
   * @param array $form
   *   The form object.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   */
  public function toMapping(array &$form, FormStateInterface $form_state): void {
    $form_state->setRedirectUrl($this->entity->toUrl('mapping'));
  }

  /**
   * AJAX callback to update the field wrapper.
   *
   * @param array<string, mixed> $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return mixed
   *   The form element wrapper.
   */
  public function updateFields(array &$form, FormStateInterface $form_state) {
    return $form['fields_wrapper'];
  }

  /**
   * Checks if a mapping ID already exists.
   */
  public function exist($id): bool {
    $entity = $this->entityTypeManager->getStorage('field_updater_type')->load($id);
    return (bool) $entity;
  }

  /**
   * Page title callback.
   *
   * @return string
   *   The title of the mapping page.
   */
  public function title(FieldUpdaterTypeInterface $field_updater_type): string {
    return (string) $this->t('Edit @label', ['@label' => $field_updater_type->label()]);
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state): int {
    /** @var \Drupal\field_updater_service\Entity\FieldUpdaterType $entity */
    $entity = $this->entity;
    $values = $form_state->getValues();
    $values['target_bundles'] = !empty($values['target_bundles']) ? array_keys(array_filter($values['target_bundles'])) : [];
    $changed = [];
    if (!$entity->isNew()) {
      /** @var \Drupal\field_updater_service\Entity\FieldUpdaterType $original_entity */
      $original_entity = $this->entityTypeManager->getStorage('field_updater_type')->loadUnchanged($entity->id());
      $existing_values = [
        'entity_type' => $original_entity->entity_type,
        'bundle' => $original_entity->bundle,
        'field_target' => $original_entity->field_target,
        'field_source' => $original_entity->field_source,
        'target_type' => $original_entity->target_type,
        'target_bundles' => $original_entity->target_bundles ?? [],
      ];
      // Ensure $values only includes the keys we care about to avoid irrelevant
      // differences.
      $submitted_values = array_intersect_key($values, $existing_values);
      $submitted_values['target_bundles'] = is_array($submitted_values['target_bundles']) ? implode(',', $submitted_values['target_bundles']) : '';
      $existing_values['target_bundles'] = is_array($existing_values['target_bundles']) ? implode(',', $existing_values['target_bundles']) : '';
      $changed = array_diff_assoc($submitted_values, $existing_values);
    }

    $entity->set('entity_type', $values['entity_type']);
    $entity->set('bundle', $values['bundle']);
    $entity->set('field_target', $values['field_target']);
    $entity->set('field_source', $values['field_source'] ?? NULL);
    $entity->set('target_type', $values['target_type'] ?? NULL);
    $entity->set('target_bundles', $values['target_bundles']);
    if (!empty($changed)) {
      $entity->setMappings([]);
    }

    $status = parent::save($form, $form_state);

    $this->messenger()->addStatus(
      $status == SAVED_NEW
        ? $this->t('Created the %label Field updater.', ['%label' => $entity->label()])
        : $this->t('Updated the %label Field updater.', ['%label' => $entity->label()])
    );

    $form_state->setRedirect('entity.field_updater_type.collection');
    return $status;
  }

  /**
   * Helper function to return entity types and bundles.
   *
   * @return array
   *   An array of entity type ids with extra bundle info.
   */
  protected function getEntityTypesWithBundles(): array {
    $types_with_bundles = [];
    foreach ($this->entityTypeManager->getDefinitions() as $entity_name => $entity_definition) {
      if ($entity_definition->entityClassImplements(FieldableEntityInterface::class)) {
        $types_with_bundles[$entity_name] = [
          'machine_name' => $entity_name,
          'label' => $entity_definition->getLabel(),
          'bundles' => [],
        ];
      }
    }
    foreach ($types_with_bundles as $entity_type_key => &$entity_type) {
      foreach ($this->entityTypeBundleInfo->getBundleInfo($entity_type_key) as $bundle => $bundle_info) {
        $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type_key, $bundle);
        $bundle_fields = [];
        foreach ($field_definitions as $field_name => $definition) {
          if (!$definition->getFieldStorageDefinition()->isBaseField()) {
            $bundle_fields[] = $field_name;
          }
        }
        if (count($bundle_fields) >= 1) {
          $entity_type['bundles'][$bundle] = $bundle_info['label'];
        }
      }
      if (count($entity_type['bundles']) === 0) {
        unset($types_with_bundles[$entity_type_key]);
      }
    }
    return $types_with_bundles;
  }

}
