<?php

declare(strict_types=1);

namespace Drupal\field_updater_service\Form;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\FocusFirstCommand;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Config\Entity\ConfigEntityStorageInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\field_updater_service\FieldUpdaterTypeInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a form for mapping settings.
 */
class MappingForm extends FormBase {

  /**
   * The field updater type.
   *
   * @var \Drupal\field_updater_service\FieldUpdaterTypeInterface|null
   */
  protected ?FieldUpdaterTypeInterface $fieldUpdater;

  /**
   * The field updater type storage.
   *
   * @var \Drupal\Core\Config\Entity\ConfigEntityStorageInterface
   */
  protected ConfigEntityStorageInterface $fieldUpdaterStorage;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected EntityFieldManagerInterface $entityFieldManager;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected ModuleHandlerInterface $moduleHandler;

  /**
   * The token entity mapper service.
   *
   * @var \Drupal\token\TokenEntityMapperInterface
   */
  protected $tokenEntityMapper;

  /**
   * The renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected RendererInterface $renderer;

  /**
   * The mappings for this field updater type.
   *
   * @var array
   */
  protected array $mappings = [];

  /**
   * The mapping targets for this field type.
   *
   * @var array
   */
  protected array $targets = [];

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->fieldUpdaterStorage = $container->get('entity_type.manager')->getStorage('field_updater_type');
    $instance->entityFieldManager = $container->get('entity_field.manager');
    $instance->moduleHandler = $container->get('module_handler');
    $instance->renderer = $container->get('renderer');
    $instance->tokenEntityMapper = $container->get(
      'token.entity_mapper',
      ContainerInterface::NULL_ON_INVALID_REFERENCE
    );

    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'field_updater_mapping_form';
  }

  /**
   * {@inheritdoc}
   *
   * @return array<string, mixed>
   *   The form.
   */
  public function buildForm(array $form, FormStateInterface $form_state, ?FieldUpdaterTypeInterface $field_updater_type = NULL): array {
    $field_updater = $this->fieldUpdater = $field_updater_type;
    $entity_type_id = $field_updater->getUpdaterEntityTypeId();
    $bundle = $field_updater->getUpdaterBundle();
    $field_target = $field_updater->get('field_target');
    $field_source = $field_updater->get('field_source');
    $mappings = $this->fieldUpdater->getMappings();
    $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);
    $target_definition = $field_definitions[$field_target];
    /** @var \Drupal\Core\Field\FieldDefinitionInterface[] $eligible_mappings */
    $eligible_mappings = [];
    $source_definition = $field_definitions[$field_source] ?? NULL;
    if (!empty($source_definition)) {
      if (in_array($source_definition->getType(), ['entity_reference_revisions', 'entity_reference'])) {
        $target_type = $field_updater->get('target_type');
        $target_bundles = $field_updater->get('target_bundles');
        // Get fields for all bundle types.
        foreach ($target_bundles as $target_bundle) {
          $eligible_mappings = array_merge(
            $eligible_mappings,
            $this->entityFieldManager->getFieldDefinitions($target_type, $target_bundle)
          );
        }
      }
      else {
        $eligible_mappings[$source_definition->getName()] = $source_definition;
      }
    }
    else {
      $eligible_mappings = array_filter($field_definitions, function ($field_name) use ($field_target) {
        return $field_name !== $field_target;
      }, ARRAY_FILTER_USE_KEY);
    }

    $form['#tree'] = TRUE;
    $form['description'] = [
      '#type' => 'markup',
      '#markup' => 'At least 1 target field must be mapped to a source field. Empty fields thereafter will be skipped when this updater is processed.',
    ];

    $header = [
      'target' => $this->t('Target'),
      'source_field' => $this->t('Field source'),
      'source_property' => $this->t('Value source'),
    ];

    $form['mappings'] = [
      '#type' => 'table',
      '#header' => $header,
    ];
    $target_storage = $target_definition->getFieldStorageDefinition();
    $target_schema = $target_storage->getSchema();
    foreach ($target_storage->getPropertyDefinitions() as $property => $property_definition) {
      // Skip computed properties.
      if ($property_definition->isComputed()) {
        continue;
      }
      $data_type = $property_definition->getDataType();
      $target_column = $target_schema['columns'][$property];
      $target_fields = [];
      $target_options = [];
      $strict_matches = [
        'boolean',
        'datetime_iso8601',
        'filter_format',
        'timestamp',
      ];

      // Static default options available for all properties.
      $default_options = [
        '_default_null' => $this->t('NULL'),
        '_default_value' => $this->t('Custom value'),
      ];

      // Restrict cases that don't make sense to have a default value option
      // until we can provide a viable widget or validation.
      if ($data_type === 'datetime_iso8601') {
        unset($default_options['_default_value']);
      }

      $eligible_columns = [];
      foreach ($eligible_mappings as $name => $eligible_mapping) {
        if ($eligible_mapping->getType() === 'entity_reference_revisions') {
          continue;
        }

        $mapping_storage = $eligible_mapping->getFieldStorageDefinition();
        $mapping_columns = $mapping_storage->getSchema()['columns'];
        $mapping_properties = $mapping_storage->getPropertyDefinitions();
        $has_matching_type = FALSE;

        foreach ($mapping_properties as $mapping_property_name => $mapping_property) {
          if ($mapping_property->isComputed()) {
            continue;
          }
          $mapping_data_type = $mapping_property->getDataType();
          $mapping_column = $mapping_columns[$mapping_property_name] ?? NULL;

          // Match on data type (strict matches only).
          if ($data_type === $mapping_data_type && in_array($data_type, $strict_matches)) {
            $has_matching_type = TRUE;
            $eligible_columns[$name][$mapping_property_name] = $mapping_property_name;
            break;
          }

          // Match with a schema type and additional constraints.
          if ($mapping_column && $mapping_column['type'] === $target_column['type']) {
            // Check size if present.
            if (isset($mapping_column['size']) || isset($target_column['size'])) {
              if (($mapping_column['size'] ?? NULL) === ($target_column['size'] ?? NULL)) {
                $has_matching_type = TRUE;
                $eligible_columns[$name][$mapping_property_name] = $mapping_property_name;
                continue;
              }
              continue;
            }
            // Check length if present.
            if (isset($mapping_column['length']) && isset($target_column['length'])) {
              if ($target_column['length'] >= $mapping_column['length']) {
                $has_matching_type = TRUE;
                $eligible_columns[$name][$mapping_property_name] = $mapping_property_name;
                continue;
              }
              continue;
            }
            // No size/length constraints, schema type match is enough.
            $has_matching_type = TRUE;
            $eligible_columns[$name][$mapping_property_name] = $mapping_property_name;
          }
        }

        if ($has_matching_type) {
          $label = $this->t('@field_label (@field_name)', [
            '@field_label' => $eligible_mapping->getLabel(),
            '@field_name' => $name,
          ]);
          $target_fields[$name] = $label;
        }
      }
      if (!empty($target_fields)) {
        $target_options['Fields'] = $target_fields;
      }
      $target_options += ['Default values' => $default_options];

      $form['mappings'][$property] = [
        '#type' => 'container',
      ];
      $form['mappings'][$property]['target'] = [
        '#type' => 'markup',
        '#markup' => $this->t('@field_name.@property', ['@field_name' => $field_target, '@property' => $property]),
      ];
      $form['mappings'][$property]['source_field'] = [
        '#type' => 'select',
        '#title' => $this->t('Source field for target @property', ['@property' => $property]),
        '#title_display' => 'invisible',
        '#options' => $target_options,
        '#empty_option' => $this->t('- None -'),
        '#default_value' => $mappings[$property]['source_field'] ?? NULL,
        '#ajax' => [
          'callback' => [$this, 'updateSourcePropertyOptions'],
          'wrapper' => "edit-mappings-$property-source-property-wrapper",
          'event' => 'change',
        ],
      ];

      // Get the selected source field value from form state or default
      // mappings.
      $selected_source_field = $form_state->getValue(['mappings', $property, 'source_field']) ?? ($mappings[$property]['source_field'] ?? NULL);
      $source_property_options = [];
      if (!empty($selected_source_field) && isset($eligible_columns[$selected_source_field])) {
        $source_property_options = $eligible_columns[$selected_source_field];
      }

      $property_required = !empty($selected_source_field) && $selected_source_field !== '_default_null';

      $form['mappings'][$property]['source_property'] = [
        '#type' => 'select',
        '#title' => $this->t('Source property for target @property', [
          '@property' => $property,
        ]),
        '#title_display' => 'invisible',
        '#options' => $source_property_options,
        '#empty_option' => $this->t('- Select -'),
        '#required' => $property_required,
        '#default_value' => $mappings[$property]['source_property'] ?? NULL,
        '#prefix' => '<div id="edit-mappings-' . $property . '-source-property-wrapper">',
        '#suffix' => '</div>',
      ];
      if (empty($selected_source_field)) {
        $form['mappings'][$property]['source_property']['#type'] = 'hidden';
        $form['mappings'][$property]['source_property']['#value'] = NULL;
        unset($form['mappings'][$property]['source_property']['#options']);
      }
      // Alter the default value widgets.
      elseif ($selected_source_field === '_default_value') {
        unset($form['mappings'][$property]['source_property']['#options']);
        $tokens_supported = FALSE;
        if ($data_type === 'boolean') {
          $form['mappings'][$property]['source_property']['#type'] = 'radios';
          $form['mappings'][$property]['source_property']['#options'] = [
            1 => $this->t('True'),
            0 => $this->t('False'),
          ];
        }
        elseif ($target_column['type'] === 'int') {
          $form['mappings'][$property]['source_property']['#type'] = 'number';
          $unsigned = $target_column['unsigned'] ?? FALSE;
          if ($unsigned) {
            $form['mappings'][$property]['source_property']['#min'] = 0;
          }
        }
        elseif ($target_column['type'] === 'text') {
          $form['mappings'][$property]['source_property']['#type'] = 'textarea';
          $form['mappings'][$property]['source_property']['#rows'] = 3;
          $tokens_supported = TRUE;
        }
        else {
          $form['mappings'][$property]['source_property']['#type'] = 'textfield';
          if (isset($target_column['length'])) {
            $form['mappings'][$property]['source_property']['#maxlength'] = (int) $target_column['length'];
          }
          $tokens_supported = TRUE;
        }
        // Add token browser.
        if ($tokens_supported && $this->moduleHandler->moduleExists('token')) {
          $token_type = $this->tokenEntityMapper->getTokenTypeForEntityType($entity_type_id);
          if ($token_type) {
            $token_tree = [
              '#theme' => 'token_tree_link',
              '#token_types' => [$token_type],
            ];
            $form['mappings'][$property]['source_property']['#description'] = $this->renderer->renderInIsolation($token_tree);
          }
        }
      }
      elseif ($selected_source_field === '_default_null') {
        $form['mappings'][$property]['source_property']['#type'] = 'value';
        $form['mappings'][$property]['source_property']['#value'] = NULL;
        unset($form['mappings'][$property]['source_property']['#options']);
      }
    }

    $form['actions'] = ['#type' => 'actions'];
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save'),
      '#button_type' => 'primary',
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): void {
    // Validate mappings.
    $mappings = $form_state->getValue('mappings');
    $filtered_mappings = array_filter($mappings, function ($mapping) {
      return !empty($mapping['source_field']) && $mapping['source_field'] !== '_default_null';
    });
    if (empty($filtered_mappings)) {
      $form_state->setErrorByName('mappings', $this->t('At least one field source is required.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $mappings = $form_state->getValue('mappings');
    $this->fieldUpdater->setMappings($mappings);
    $this->fieldUpdater->save();
    $this->messenger()->addMessage($this->t('Updated mappings for the <em>@label</em> field updater.', [
      '@label' => $this->fieldUpdater->label(),
    ]));
  }

  /**
   * Page title callback.
   *
   * @return string
   *   The title of the mapping page.
   */
  public function mappingTitle(FieldUpdaterTypeInterface $field_updater_type): string {
    return (string) $this->t('Mappings @label', ['@label' => $field_updater_type->label()]);
  }

  /**
   * AJAX callback to update the source_property options.
   *
   * @param array<string, mixed> $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   The ajax response.
   */
  public function updateSourcePropertyOptions(array &$form, FormStateInterface $form_state): AjaxResponse {
    $triggering_element = $form_state->getTriggeringElement();
    $wrapper = $triggering_element['#ajax']['wrapper'];
    $sliced_parents = array_slice($triggering_element['#array_parents'], 0, -1);
    $updated_element = NestedArray::getValue($form, [...$sliced_parents, 'source_property']);
    $response = new AjaxResponse();
    $response->addCommand(new ReplaceCommand('#' . $wrapper, $updated_element));
    if ($updated_element['#type'] !== 'value' && $updated_element['#type'] !== 'hidden') {
      $response->addCommand(new FocusFirstCommand('#' . $wrapper));
    }

    return $response;
  }

}
